import { Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { EventEmitter } from 'events';
import { FindManyOptions, In, Like, MoreThanOrEqual, Repository } from 'typeorm';

import { JobsService } from '../jobs/jobs.service';
import { UserService } from '../user/user.service';
import { QueryBuilder } from '../utility/QueryBuilder';

import { EventAttendance, EventAttendee } from './attendee.entity';
import { Event } from './event.entity';

enum Events {
  EVENT_ADDED = 'event_added',
  ATTENDANCE = 'event_attended',
  REMINDER = 'event_reminder'
}

@Injectable()
export class EventsService {

  private readonly eventEmitter = new EventEmitter();

  constructor(
    @InjectRepository(Event)
    public readonly repository: Repository<Event>, // public for testing
    @InjectRepository(EventAttendee)
    private readonly attendeeRepository: Repository<EventAttendee>,
    private readonly userService: UserService,
    private readonly jobService: JobsService) { }

  public async create(title: string, location: string, description: string, dateStart: number, dateModified: number, dateEnd: number, dateCreated: number): Promise<Event> {
    const entity = await this.repository.save({ title, location, description, dateStart, dateModified, dateEnd, dateCreated });
    this.eventEmitter.emit(Events.EVENT_ADDED, entity);
    return entity;
  }

  public onEvent(logic: (event: Event) => Promise<void>) {
    this.eventEmitter.on(Events.EVENT_ADDED, logic);
  }

  public async getAll(page: number, limit: number, query: ParameterDecorator): Promise<{ results: Event[], total: number }> {
    const { 0: events, 1: count } = await this.repository
      .createQueryBuilder('events')
      .skip(page || 0)
      .orderBy('dateStart', 'ASC')
      .take(limit && limit >= 0 ? limit : 0)
      .select()
      .where(QueryBuilder.buildTypeORMQuery(query))
      .andWhere('events.dateStart >= :date', { date: Math.round(new Date().getTime() / 1000) })
      .getManyAndCount();

    return { results: events, total: count };
  }

  public async getEvent(uuid: string): Promise<Event> {
    return await this.repository.findOneOrFail(uuid);
  }

  public async setReminder(eventId: string, userId: string, date: Date) {
    try {
      await this.clearReminder(eventId, userId);
    } catch (err) { }
    await this.jobService.add(Events.REMINDER, date, [eventId, userId]);
  }

  public async getReminder(eventId: string, userId: string) {
    return await this.jobService.one({ trigger: Events.REMINDER, args: JSON.stringify([eventId, userId]) });
  }

  public async clearReminder(eventId: string, userId: string) {
    const reminder = await this.getReminder(eventId, userId);
    await this.jobService.remove([reminder]);
  }

  public onReminder(logic: (eventId: string, userId: string) => Promise<void>) {
    this.jobService.on(Events.REMINDER, async (args: string[]) => {
      const [eventId, userId] = args;
      await logic(eventId, userId);
    });
  }

  public async setAttendance(eventId: string, userId: string, attendance: EventAttendance, preventEmit = false) {
    let entity = await this.getAttendance(eventId, userId);
    if (!entity) {
      const user = await this.userService.one({ id: userId });
      entity = this.attendeeRepository.create({
        user,
        event: { id: eventId }
      });
    }

    entity.status = attendance;
    entity = await this.attendeeRepository.save(entity);
    if (!preventEmit) { this.eventEmitter.emit(Events.ATTENDANCE, entity); }
    return entity;
  }

  public onAttendee(logic: (attendee: EventAttendee) => Promise<void>) {
    this.eventEmitter.on(Events.ATTENDANCE, logic);
  }

  public async getAttendance(eventId: string, userId: string): Promise<EventAttendee> {
    return await this.attendeeRepository.findOne({ event: { id: eventId }, user: { id: userId } });
  }

  public async getManyAttendance(eventIds: string[], userId: string): Promise<EventAttendee[]> {
    return await this.attendeeRepository.find({
      where: { event: In(eventIds), user: { id: userId } }
    });
  }

  public async findAttendance(conditions: FindManyOptions<EventAttendee>): Promise<EventAttendee[]> {
    return await this.attendeeRepository.find(conditions);
  }
}
