import { forwardRef, Inject, Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Connection, FindConditions, Repository } from 'typeorm';

import { User } from '../user/user.entity';
import { UserMeta } from '../user/usermeta.entity';
import { UserMetaEnum } from '../user/usermeta.enum';
import { UserMetaService } from '../user/usermeta.service';

import { Organiser } from './organiser.entity';

@Injectable()
export class OrganiserService {

    constructor(
        @InjectRepository(Organiser)
        private readonly repository: Repository<Organiser>,
        @Inject(forwardRef(() => UserMetaService))
        private readonly userMetaService: UserMetaService,
        private readonly dbConn: Connection) { }

    public async create(name: string, userId: string): Promise<Organiser> {
        let organiser: Organiser;
        await this.dbConn.transaction(async (manager) => {
            const userMetaService = new UserMetaService(manager.getRepository(UserMeta));

            // save the organiser and set its admin
            organiser = await manager.getRepository(Organiser).save({ name });
            await userMetaService.add(userId, UserMetaEnum.ORGA_ADMIN, organiser.id);
        });
        return organiser;
    }

    /**
     * @throws an error if the user doing the operation doesn't have full admin rights for the organiser
     * @param organiserId
     * @param userId
     * @param actingUserId
     */
    public async addAdmin(organiserId: string, userId: string, actingUserId: string): Promise<void> {
        if (!this.isFullAdmin(organiserId, await this.getAdmin(organiserId, actingUserId))) {
            throw new Error('Current user is not permitted to add an admin');
        }
        await this.userMetaService.add(userId, UserMetaEnum.ORGA_ADMIN, organiserId);
    }

    /**
     * @throws an error if the user doing the operation doesn't have full admin rights for the organiser,
     *   or the organiser only has one admin
     * @param organiserId
     * @param userId
     * @param actingUserId
     */
    public async removeAdmin(organiserId: string, userId: string, actingUserId: string): Promise<void> {
        const admins = await this.getAdmins(organiserId);

        if (admins.length === 1) {
            if (admins[0].user.id !== userId) { return; } // no need telling there's only one admin
            throw new Error('Removing the only admin is not permitted');
        }

        const sameUser = actingUserId === userId;
        if (sameUser) { actingUserId = ''; }

        let adminMeta: UserMeta,
            actingAdminMeta: UserMeta;

        for (const admin of admins) {
            if (adminMeta && actingAdminMeta) { break; }
            switch (admin.user.id) {
                case userId:
                adminMeta = admin;
                if (sameUser) { actingAdminMeta = admin; }
                continue;

                case actingUserId:
                actingAdminMeta = admin;
                continue;
            }
        }

        if (!(adminMeta && actingAdminMeta)) { return; }

        // check if the acting user is trusted
        if (!sameUser && !await this.isFullAdmin(organiserId, actingAdminMeta)) {
            throw new Error('The current admin account is not a full admin yet');
        }

        await this.userMetaService.remove([adminMeta]); // remove the admin
    }

    public async getAdmin(organiserId: string, userId: string): Promise<UserMeta> {
        return await this.userMetaService.repository.findOneOrFail({
            where: { key: UserMetaEnum.ORGA_ADMIN, value: organiserId, user: { id: userId } }
        });
    }

    public async getAdmins(organiserId: string): Promise<UserMeta[]> {
        return await this.userMetaService.repository.find({
            where: { key: UserMetaEnum.ORGA_ADMIN, value: organiserId },
            order: { dateCreated: 'ASC' }
        });
    }

    /**
     * Get the organiser ids a user manages. It also gives out the total count of matched organisers.
     * @param id User id
     */
    public async getManagedOrganizers(id: string): Promise<[string[], number]> {
        return await this.userMetaService.getManagedOrganizers(id);
    }

    private async isFullAdmin(organiserId: string, userMeta: UserMeta): Promise<boolean> {
        const week = 7 * 24 * 60 * 60;
        return Date.now() >= (userMeta.dateCreated.valueOf() + week) // admin is over a week old
            || userMeta.id === (await this.getAdmins(organiserId))[0].id // oldest admin
            || ((await this.one({ id: organiserId })).dateCreated.valueOf() + week) > Date.now(); // new organiser, any change allowed
    }

    /**
     * @throws an error if no organiser could match the @param conditions.
     */
    public async one(conditions: FindConditions<Organiser>): Promise<Organiser> {
        return this.repository.findOneOrFail(conditions);
    }

    public async find(conditions: FindConditions<Organiser>): Promise<Organiser[]> {
        return this.repository.find(conditions);
    }

    public async exists(conditions: FindConditions<Organiser>): Promise<boolean> {
        try {
            await this.repository.findOneOrFail(conditions, { select: ['id']  });
            return true;
        } catch (error) {
            return false;
        }
    }

    /**
     * @throws an error if the user doing the operation doesn't have full admin rights for the organiser.
     */
    public async remove(organiser: Organiser, actingUserId: string): Promise<void> {
        const admin = await this.getAdmin(organiser.id, actingUserId);
        if (!await this.isFullAdmin(organiser.id, admin)) {
            throw new Error('The current admin account is not permitted to run this operation');
        }

        const admins = await this.getAdmins(organiser.id);
        await this.repository.remove(organiser);
        await this.userMetaService.remove(admins);
    }

    public async subscribe(organiserId: string, userId: string): Promise<void> {
        // save the organiser and set its admin
        await this.userMetaService.add(userId, UserMetaEnum.SUB, organiserId);
    }

    public async unsubscribe(organiserId: string, userId: string): Promise<void> {
        // save the organiser and set its admin
        const entity = await this.userMetaService.repository.find({ user: { id: userId }, key: UserMetaEnum.SUB, value: organiserId });
        await this.userMetaService.remove(entity);
    }

    /**
     * Get the user accounts who have subscribed to the organiser.
     * @param organiserId
     * @param skip
     * @param take
     */
    public async getSubscribers(organiserId: string, skip: number, take?: number): Promise<User[]> {
        const subs = await this.userMetaService.repository.find({ where: { key: UserMetaEnum.SUB, value: organiserId }, skip, take });
        return subs.map((userMeta) => {
            return userMeta.user;
        });
    }
}
