import { Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import * as generate from 'nanoid/generate';
import { MoreThanOrEqual, Repository } from 'typeorm';

import { Session } from './session.entity';
import { User } from './user.entity';

const HOUR = 60 * 60 * 1000;

@Injectable()
export class SessionService {

    constructor(
        @InjectRepository(Session)
        private readonly repository: Repository<Session>) {
        // Start the cleanup job.
        setInterval(this.removeExpired.bind(this), HOUR);
    }

    public async one(id: string): Promise<Session> {
        return this.repository.findOneOrFail(id);
    }

    public async create(user: Pick<User, 'id'>): Promise<Session> {
        const alphabet = '23456789ABCDEFGHJKLMNPQRSTUVWXYZ';

        let maxTries = 3;
        do {
            const id = generate(alphabet, 16);
            const collision = await this.repository.findOne(id);
            if ('undefined' !== typeof collision) { continue; } // collision, regenerate

            const session = this.repository.create({
                id,
                user,
                dateStarted: Date.now(),
                userAgent: ''
            });

            try {
                return await this.repository.save(session);
            } catch (error) {
                maxTries--;
                if (maxTries > 0) { continue; }
                throw new Error('Couldn\'t create a session' + error);
            }
        } while (maxTries > 0);
    }

    public async update(session: Session): Promise<Session> {
        return await this.repository.save(session);
    }

    /**
     * A session is valid if it is either less than an hour old,
     * or it is less than a year old (cookie expiration)
     * and it contains the user agent of the connected client.
     * @param session
     */
    public isValid(session: Session): boolean {
        if (!('dateStarted' in session)) { return false; }

        const timeDiff = Date.now() - session.dateStarted;

        // less than an hour old
        if (timeDiff < HOUR) {
            return true;
        }

        // less than a year old and contains a user agent
        if (session.userAgent.length && timeDiff < (HOUR * 24 * 365)) {
            return true;
        }
        return false;
    }

    public async remove(session: Session) {
        await this.repository.remove(session);
    }

    /**
     * Clean up the expired sessions.
     */
    public async removeExpired() {
        const now = Date.now();

        // remove any session older than an hour without a user agent
        let expired = await this.repository.find({
            userAgent: '',
            dateStarted: MoreThanOrEqual(now + HOUR)
        }) || [];
        await this.repository.remove(expired);

        // remove any session that is older than a year
        expired = await this.repository.find({
            dateStarted: MoreThanOrEqual(now + (HOUR * 24 * 365))
        }) || [];
        await this.repository.remove(expired);
    }
}
