import { Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Connection, FindConditions, Repository } from 'typeorm';

import { Session } from './session.entity';
import { User } from './user.entity';

@Injectable()
export class UserService {

    constructor(
        @InjectRepository(User)
        private readonly repository: Repository<User>,
        private readonly dbConn: Connection) {
    }

    /**
     * @throws an Error if no match was made.
     * @param conditions
     */
    public async one(conditions: FindConditions<User>): Promise<User> {
        return this.repository.findOneOrFail(conditions);
    }

    public async find(conditions: FindConditions<User>): Promise<User[]> {
        return this.repository.find(conditions);
    }

    public async exists(conditions: FindConditions<User>): Promise<boolean> {
        try {
            await this.repository.findOneOrFail(conditions, { select: ['id'] });
            return true;
        } catch (error) {
            return false;
        }
    }

    /**
     * Create a user.
     * @throws an Error if a user with the given telegramId already exists
     * @param name
     * @param telegramId
     */
    public async create(name: string, telegramId: string): Promise<User> {
        try {
            await this.one({ telegramId });
        } catch (error) {
            // user doesn't exist
            return this.repository.save({ name, telegramId });
        }
        throw new Error('User already exists');
    }

    public async update(id: string, name: string): Promise<User> {
        return await this.repository.save({ id, name });
    }

    /**
     * Permanently delete the given user and its related entities.
     * @param user
     */
    public async remove(user: User): Promise<void> {
        await this.dbConn.transaction(async (manager) => {
            // remove all the user's sessions
            const sessRepo = manager.getRepository(Session);
            const sessions = await sessRepo.find({ user });
            await sessRepo.remove(sessions);

            // remove the user
            await manager.getRepository(User).remove(user);
        });
    }

    public async updateOne(user: User): Promise<User> {
        return await this.repository.save(user);
    }
}
