import { Injectable } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Repository } from 'typeorm';

import { UserMeta } from './usermeta.entity';
import { UserMetaEnum } from './usermeta.enum';

@Injectable()
export class UserMetaService {

    constructor(
        @InjectRepository(UserMeta)
        public readonly repository: Repository<UserMeta>) {
    }

    /**
     * @throws an error if the combination of id, key and value already exists
     * @param id User id
     * @param key Meta key
     * @param value Meta value
     */
    public async add(id: string, key: UserMetaEnum | string, value: string): Promise<UserMeta> {
        return await this.repository.save({
            key,
            value,
            user: { id }
        });
    }

    public async find(id: string, key?: UserMetaEnum | string): Promise<UserMeta[]> {
        return await this.repository.find({ user: { id }, key });
    }

    public async clear(id: string): Promise<void> {
        const entities = await this.repository.find({ user: { id }});
        await this.remove(entities);
    }

    public async remove(entities: UserMeta[]): Promise<void> {
        await this.repository.remove(entities);
    }

    /**
     * Get the organisers a user is subscribed to.
     * It also gives out the total subscription count.
     * @param id User id
     * @param skip
     * @param take
     */
    public async getSubscriptions(id: string, skip: number, take?: number): Promise<[string[], number]> {
        const entities = await this.repository.findAndCount({
            where: { user: { id }, key: UserMetaEnum.SUB },
            skip,
            take
        });

        return [
            entities[0].map((meta) => meta.value),
            entities[1]
        ];
    }

    /**
     * Get the organiser ids a user manages.
     * It also gives out the total count of matched organisers.
     * @param id User id
     */
    public async getManagedOrganizers(id: string): Promise<[string[], number]> {
        const entities = await this.repository.findAndCount({
            where: { user: { id }, key: UserMetaEnum.ORGA_ADMIN }
        });

        return [
            entities[0].map((meta) => meta.value),
            entities[1]
        ];
    }
}
